﻿// The following script demonstrates how to
// send native index/vertex pointer and texture resource
// to the native rendering plugin.  This script will
// also send the object and camera's transform data to
// the native plugin so it doesn't need to be recalculated.

using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.Rendering;
using System.Runtime.InteropServices;

public class CustomRenderPluginExample : MonoBehaviour
{
    public Camera m_Camera = null;
    public MeshFilter m_MeshFilter = null;
    public Texture2D m_MeshTexture = null;
    public Transform m_WorldObjectTransform = null;

    // SetupResources and GetRenderEventFunc are the two methods found in our native rendering plugin.
    [DllImport("NativeRenderingPlugin")]
    private static extern void SetupResources(System.IntPtr nativeIndexBuffer, System.IntPtr nativeVertexBuffer, System.IntPtr nativeTextureResource);

    // GetRenderEventFunc will return a native function pointer to our native DoRenderEvent method
    // which will ultimately intialize our native graphics state object and handle the native rendering of
    // our mesh.
    [DllImport("NativeRenderingPlugin")]
    private static extern System.IntPtr GetRenderEventFunc();

    CommandBuffer cb = null;
    System.IntPtr nativeRenderingDataPtr;

    enum CustomRenderEvent
    {
        // 3245 is a random number I made up.
        // I figured it could be useful to send an event id
        // to the native plugin which corresponds to when
        // the mesh is being rendered in the render pipeline.
        AfterForwardOpaque = 3245,
    }

    // These are the matrices we are going to send from Unity to
    // the native rendering plugin.
    [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
    struct NativeRenderingData
    {
        public const int Size =
            sizeof(float) * 16 +
            sizeof(float) * 16 +
            sizeof(float) * 16;

        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
        public float[] localToWorldMatrix;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
        public float[] worldToViewMatrix;
        [MarshalAs(UnmanagedType.ByValArray, SizeConst = 16)]
        public float[] viewToProjectionMatrix;
    };

    NativeRenderingData nativeRenderingData = new NativeRenderingData();

    // Use this for initialization
    void Awake ()
    {
        Mesh mesh = m_MeshFilter.sharedMesh;

        System.IntPtr nativeIndexBufferPtr = mesh.GetNativeIndexBufferPtr();
        System.IntPtr nativeVertexBufferPtr = mesh.GetNativeVertexBufferPtr(0);
        System.IntPtr nativeTexturePtr = m_MeshTexture.GetNativeTexturePtr();

        // Pass the native graphics API index/vertex/texture
        // pointers to the native rendering plugin.
        SetupResources(nativeIndexBufferPtr, nativeVertexBufferPtr, nativeTexturePtr);

        nativeRenderingData = new NativeRenderingData();
        nativeRenderingData.localToWorldMatrix = new float[16];
        nativeRenderingData.worldToViewMatrix = new float[16];
        nativeRenderingData.viewToProjectionMatrix = new float[16];

        // Allocate unmanaged memory to hold all three of our matrices that we
        // weish to send to our native rendering plugin.
        nativeRenderingDataPtr = Marshal.AllocHGlobal(NativeRenderingData.Size);

        cb = new CommandBuffer();
    }

    // There might be a better way to do this but
    // I want to marshal the C# Matrix4x4 data to C++
    // without any marshaling issues.
    void MatrixToFloatArray(Matrix4x4 m, ref float[] outputFloatArray)
    {
        outputFloatArray[0] = m.m00;
        outputFloatArray[1] = m.m01;
        outputFloatArray[2] = m.m02;
        outputFloatArray[3] = m.m03;

        outputFloatArray[4] = m.m10;
        outputFloatArray[5] = m.m11;
        outputFloatArray[6] = m.m12;
        outputFloatArray[7] = m.m13;

        outputFloatArray[8] = m.m20;
        outputFloatArray[9] = m.m21;
        outputFloatArray[10] = m.m22;
        outputFloatArray[11] = m.m23;

        outputFloatArray[12] = m.m30;
        outputFloatArray[13] = m.m31;
        outputFloatArray[14] = m.m32;
        outputFloatArray[15] = m.m33;
    }

    private void OnPreRender()
    {
        // If we didn't care about updating the matrix data dynamically,
        // we could just attach a single command buffer to our camera in the Start method.
        // However because we want our model to rotate, we need to update the matrix data in
        // the native rendering plugin.
        m_Camera.RemoveCommandBuffer(CameraEvent.AfterForwardOpaque, cb);
        cb.Release();

        // Don't pass the camera's projection matrix directly into the plugin.
        // We need to calculate the GPU ready version of the projection matrix
        // which enable the projection matrix to play nicely with reverse z depth
        // testing and inverted uv coordinates to go from opengl to d3d11.
		// The projection matrix will also be modified such that the normalized device
        // coordinates for the z component will be converted from -1 to 1 (OpenGL) to 0 to 1 (DX11).
        Matrix4x4 projectionMatrix = m_Camera.projectionMatrix;
        projectionMatrix = GL.GetGPUProjectionMatrix(projectionMatrix, true);

        MatrixToFloatArray(m_WorldObjectTransform.localToWorldMatrix, ref nativeRenderingData.localToWorldMatrix);
        MatrixToFloatArray(m_Camera.worldToCameraMatrix, ref nativeRenderingData.worldToViewMatrix);
        MatrixToFloatArray(projectionMatrix, ref nativeRenderingData.viewToProjectionMatrix);

        // Copy our managed nativeRenderingData into the unmanaged memory pointed to by
        // the nativeRenderingDataPtr pointer.
        Marshal.StructureToPtr(nativeRenderingData, nativeRenderingDataPtr, true);

        cb = new CommandBuffer();
        cb.name = "Native Rendering Plugin";
        // The IssuePluginEventAndData will ensure our native rendering function is executed and receives
        // our custom rendering data when the command buffer is executed on the render thread.
        cb.IssuePluginEventAndData(GetRenderEventFunc(), (int)CustomRenderEvent.AfterForwardOpaque, nativeRenderingDataPtr);
        m_Camera.AddCommandBuffer(CameraEvent.AfterForwardOpaque, cb);
    }

    private void OnDestroy()
    {
        // Be a good Citizen and clean up after yourself.  =D
        cb.Release();
        Marshal.FreeHGlobal(nativeRenderingDataPtr);
        nativeRenderingDataPtr = System.IntPtr.Zero;
    }
}
