#include "UnityRenderLogic.h"
#include "VertexShader.h"
#include "PixelShader.h"

// The Project will run a custom Pre-Build event
// which will compile our shaders and statically link them
// into the project.  The compiled byte code can be safely
// referenced via vertexShaderByteCode and pixelShaderByteCode
// which comes from the VertexShader.h and PixelShader.h
// header files.  Those header files are generated by an external
// python script I wrote so make sure you have python installed.
// The custom shader compiler batch file can be located in
// Assets/compile_all_shaders.bat
// The arguments are "-r" for release and "-d" for debug.
// If you compile a debug build, shader symbols will be included.
extern unsigned char vertexShaderByteCode[];
extern unsigned char pixelShaderByteCode[];

void UnityRenderLogic::Initialize(IUnityInterfaces* unityInterfaces, int eventId, int* data)
{
    if (eventId != AfterForwardOpaque)
        return;

    IUnityGraphicsD3D11* d3d11 = unityInterfaces->Get<IUnityGraphicsD3D11>();
    dev = d3d11->GetDevice();
    dev->GetImmediateContext(&devcon);

    InitializeGraphicPipeline(dev, devcon);
}

void UnityRenderLogic::Shutdown(IUnityInterfaces* unityInterfaces)
{

}

void UnityRenderLogic::Render(IUnityInterfaces* unityInterfaces, int eventId, int* data)
{
    if (eventId != AfterForwardOpaque)
        return;

    NativeRenderingData nativeRenderingData;
    memcpy(&nativeRenderingData, data, sizeof(nativeRenderingData));

    RenderMesh(dev, devcon, nativeRenderingData);
}

void UnityRenderLogic::SetupResources(void* indexBuffer, void* vertexBuffer, void* texture)
{
    nativeIndexBuffer = reinterpret_cast<ID3D11Buffer*>(indexBuffer);
    nativeVertexBuffer = reinterpret_cast<ID3D11Buffer*>(vertexBuffer);
    nativeTextureResource = reinterpret_cast<ID3D11Resource*>(texture);

    // It would be better for the C# script to tell us how many
    // indices there are in the mesh but for simplicity, we will
    // calculate it like this.
    D3D11_BUFFER_DESC indexBufferDesc;
    nativeIndexBuffer->GetDesc(&indexBufferDesc);
    indexCount = indexBufferDesc.ByteWidth / 2;
}

int UnityRenderLogic::CalculateVertexBufferStride(D3D11_INPUT_ELEMENT_DESC* inputElementDesc, int count)
{
    unsigned int sum = 0;

    // sizeof(DXGI_FORMAT_R32G32B32_FLOAT) = 32 bits * 3 == 96 bits
    // 96 / 8 = 12 bytes;

    for (int i = 0; i < count; ++i)
    {
        switch (inputElementDesc[i].Format)
        {
        case DXGI_FORMAT_R32G32B32_FLOAT:
            sum += 12;
            break;

        case DXGI_FORMAT_R32G32B32A32_FLOAT:
            sum += 16;
            break;

        case DXGI_FORMAT_R32G32_FLOAT:
            sum += 8;
            break;
        }
    }

    return sum;
}

void UnityRenderLogic::InitializeGraphicPipeline(ComPtr<ID3D11Device>& dev, ComPtr<ID3D11DeviceContext>& devcon)
{
    // This is meant to be an example so I am not too concerned with error handling.
    // You may wonder how I magically know all the formats and sizes that should be used
    // when creating the various pipeline state objects.  Afterall, some of these details
    // are not documented.  To figure out the various state information, I made Unity render
    // my model normally and then I took a RenderDoc capture and examined the input into
    // each of the Graphics API calls.
    unsigned int vertexByteCodeLength = sizeof(vertexShaderByteCode);

    HRESULT hr = S_OK;
    hr = dev->CreateVertexShader((const void*)vertexShaderByteCode, vertexByteCodeLength, 0, vertexShader.GetAddressOf());
    hr = dev->CreatePixelShader((const void*)pixelShaderByteCode, sizeof(pixelShaderByteCode), 0, pixelShader.GetAddressOf());

    devcon->VSSetShader(vertexShader.Get(), 0, 0);
    devcon->PSSetShader(pixelShader.Get(), 0, 0);

    D3D11_INPUT_ELEMENT_DESC inputElementDesc[4] =
    {
        { "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "TANGENT", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 24, D3D11_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD", 0, DXGI_FORMAT_R32G32_FLOAT, 0, 40, D3D11_INPUT_PER_VERTEX_DATA, 0 },
    };
    unsigned int inputElementCount = sizeof(inputElementDesc) / sizeof(D3D11_INPUT_ELEMENT_DESC);

    vertexBufferStride = CalculateVertexBufferStride(inputElementDesc, inputElementCount);

    hr = dev->CreateInputLayout(inputElementDesc, inputElementCount, (const void*)vertexShaderByteCode, vertexByteCodeLength, &vertexInputLayout);

    D3D11_SHADER_RESOURCE_VIEW_DESC srvDesc = {};
    srvDesc.ViewDimension = D3D11_SRV_DIMENSION_TEXTURE2D;
    srvDesc.Format = DXGI_FORMAT_BC1_UNORM;
    srvDesc.Texture2D.MostDetailedMip = 0;
    srvDesc.Texture2D.MipLevels = 9;
    hr = dev->CreateShaderResourceView(nativeTextureResource.Get(), &srvDesc, nativeTextureResourceView.GetAddressOf());

    D3D11_SAMPLER_DESC samplerDesc = {};
    samplerDesc.Filter = D3D11_FILTER_MIN_MAG_LINEAR_MIP_POINT;
    samplerDesc.AddressU = D3D11_TEXTURE_ADDRESS_WRAP;
    samplerDesc.AddressV = D3D11_TEXTURE_ADDRESS_WRAP;
    samplerDesc.AddressW = D3D11_TEXTURE_ADDRESS_WRAP;
    samplerDesc.MipLODBias = 0.0f;
    samplerDesc.MaxAnisotropy = 1;
    samplerDesc.ComparisonFunc = D3D11_COMPARISON_GREATER;
    hr = dev->CreateSamplerState(&samplerDesc, samplerState.GetAddressOf());

    D3D11_BUFFER_DESC constantBufferDesc = { 0 };
    constantBufferDesc.Usage = D3D11_USAGE_DEFAULT;
    constantBufferDesc.ByteWidth = sizeof(NativeRenderingData);
    constantBufferDesc.BindFlags = D3D11_BIND_CONSTANT_BUFFER;
    hr = dev->CreateBuffer(&constantBufferDesc, nullptr, constantBuffer.GetAddressOf());
}

void UnityRenderLogic::RenderMesh(ComPtr<ID3D11Device>& dev, ComPtr<ID3D11DeviceContext>& devcon, NativeRenderingData& nativeRenderingData)
{
    unsigned int strides[1] = { vertexBufferStride };
    unsigned int offsets[1] = { 0 };

    devcon->VSSetShader(vertexShader.Get(), 0, 0);
    devcon->PSSetShader(pixelShader.Get(), 0, 0);
    devcon->PSSetShaderResources(0, 1, nativeTextureResourceView.GetAddressOf());
    devcon->PSSetSamplers(0, 1, samplerState.GetAddressOf());
    devcon->VSSetConstantBuffers(0, 1, constantBuffer.GetAddressOf());
    devcon->UpdateSubresource(constantBuffer.Get(), 0, 0, &nativeRenderingData, 0, 0);

    devcon->IASetInputLayout(vertexInputLayout.Get());
    devcon->IASetIndexBuffer(nativeIndexBuffer.Get(), DXGI_FORMAT_R16_UINT, 0);
    devcon->IASetVertexBuffers(0, 1, nativeVertexBuffer.GetAddressOf(), strides, offsets);
    devcon->DrawIndexed(indexCount, 0, 0);
}